/*
  ReadPrivateChannel

  Reads the latest voltage value from a private channel on ThingSpeak every 30 seconds
  and prints to the serial port debug window.

  For an example of how to read from a public channel, see ReadChannel example

  ThingSpeak ( https://www.thingspeak.com ) is an analytic IoT platform service that allows you to aggregate, visualize and
  analyze live data streams in the cloud.

  Copyright 2017, The MathWorks, Inc.

  Documentation for the ThingSpeak Communication Library for Arduino is in the extras/documentation folder where the library was installed.
  See the accompaning licence file for licensing information.
*/

#include "FishinoThingSpeak.h"


// ***********************************************************************************************************
// This example selects the correct library to use based on the board selected under the Tools menu in the IDE.
// Yun, Ethernet shield, WiFi101 shield, esp8266, and MXR1000 are all supported.
// With Yun, the default is that you're using the Ethernet connection.
// If you're using a wi-fi 101 or ethernet shield (http://www.arduino.cc/en/Main/ArduinoWiFiShield), uncomment the corresponding line below
// ***********************************************************************************************************
//#define USE_WIFI101_SHIELD
//#define USE_ETHERNET_SHIELD

#if !defined(USE_WIFI101_SHIELD) && !defined(USE_ETHERNET_SHIELD) && !defined(ARDUINO_SAMD_MKR1000) && \
    !defined(ARDUINO_AVR_YUN) && !defined(ARDUINO_ARCH_ESP8266) && !defined(FISHINO_WIFI)
#error "Uncomment the #define for either USE_WIFI101_SHIELD or USE_ETHERNET_SHIELD"
#endif

#if defined(ARDUINO_AVR_YUN)
	#include "YunClient.h"
	YunClient client;
#elif defined(FISHINO_WIFI)
	#include <Fishino.h>
	char ssid[] = "<YOURNETWORK>";    //  your network SSID (name)
	char pass[] = "<YOURPASSWORD>";   // your network password
	int status = STATION_IDLE;
	FishinoClient  client;
#elif defined(USE_WIFI101_SHIELD) || defined(ARDUINO_SAMD_MKR1000) || defined(ARDUINO_ARCH_ESP8266)
	// Use WiFi
	#ifdef ARDUINO_ARCH_ESP8266
		#include <ESP8266WiFi.h>
	#else
		#include <SPI.h>
		#include <WiFi101.h>
	#endif
	char ssid[] = "<YOURNETWORK>";    //  your network SSID (name)
	char pass[] = "<YOURPASSWORD>";   // your network password
	int status = WL_IDLE_STATUS;
	WiFiClient  client;
#elif defined(USE_ETHERNET_SHIELD)
	// Use wired ethernet shield
	#include <SPI.h>
	#include <Ethernet.h>
	byte mac[] = { 0xDE, 0xAD, 0xBE, 0xEF, 0xFE, 0xED};
	EthernetClient client;
#endif


/*
  *****************************************************************************************
  **** Visit https://www.thingspeak.com to sign up for a free account and create
  **** a channel.  The video tutorial http://community.thingspeak.com/tutorials/thingspeak-channels/
  **** has more information. You need to change this to your channel, and your read API key
  **** IF YOU SHARE YOUR CODE WITH OTHERS, MAKE SURE YOU REMOVE YOUR READ API KEY!!
  *****************************************************************************************

  This is the ThingSpeak channel used in the write examples (31461).  It is private, and requires a
  read API key to access it.  We'll read from the first field.
*/
unsigned long myChannelNumber = 31461;
const char * myReadAPIKey = "NKX4Z5JGO4M5I18A";

void setup()
{

	Serial.begin(9600);
#ifdef ARDUINO_AVR_YUN
	Bridge.begin();
#elif defined(FISHINO_WIFI)
	Fishino.begin(ssid, pass);
#elif defined(ARDUINO_ARCH_ESP8266) || defined(USE_WIFI101_SHIELD) || defined(ARDUINO_SAMD_MKR1000)
	WiFi.begin(ssid, pass);
#else
	Ethernet.begin(mac);
#endif

	ThingSpeak.begin(client);
}

void loop()
{
	// Read the latest value from field 1 of channel 31461
	float voltage = ThingSpeak.readFloatField(myChannelNumber, 1, myReadAPIKey);

	Serial.print("Latest voltage is: ");
	Serial.print(voltage);
	Serial.println("V");

	delay(30000);
}
